/*
 * Copyright (C) 2009-2010 Sébastien Villemot <sebastien.villemot@ens.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _SSA1_SOLUTION_HH
#define _SSA1_SOLUTION_HH

#include "MMJSolution.hh"

class SSA1Solution : public MMJSolution
{
public:
  SSA1Solution(const ModelSpec &modspec_arg) throw (UnsupportedSpec);
  virtual std::string name() const;
};

#endif // !_SSA1_SOLUTION_HH
